<?php
/**
 * Survey Form: v1.5 (2014-01-16)
 * Author: Jeff Hughes (j4hughes@uwaterloo.ca)
 */

class SurveyForm {

    const FORM_PREFIX = 'sfvar_'; // prefix for name of all form elements inserted behind-the-scenes
    const OUTPUT_FILENAME = 'output.txt'; // name of the output file
    const PASSWORD_FILENAME = 'passfile.txt'; // name of the file with usernames and passwords
    const BLOCKRAND_FILENAME = 'used_conditions.txt'; // name of the file to keep track of used conditions
    
    protected $baseDir; // directory where survey exists
    protected $baseUrl; // URL where survey exists
    protected $email; // email to be used in case of errors
    protected $requiredLogin; // 'sona', 'password', or 'none'
    protected $allConditions; // array of all conditions
    protected $blockRandom; // whether to use block randomization or not
    protected $pageOrder; // customized page order

    protected $outputFile; // location of output file
    protected $passFile; // location of file with usernames and passwords
    protected $blockRandomFile; // location of file to keep track of used conditions

    protected $allVariables; // array of all variables in the study, organized by page name
    
    private $sonaID; // sona ID of participant (if $requiredLogin == 'sona')
    private $sID; // user ID of participant (if $requiredLogin == 'password')
    private $otherID; // generated ID of participant (if $requiredLogin == 'none')
    private $condition; // actual set of conditions that the participant is in
    private $lastPage; // last page submitted
    private $currentPage; // current page
    
    private $extras; // extra variables that have been passed in through $_GET
    
    
    public function __construct($baseDir, $baseUrl, $email, $requiredLogin = 'none', $allConditions, $blockRandom, $pageOrder = null) {
        @set_exception_handler(array('SurveyForm', 'exceptionHandler')); // use exceptionHandler() to deal with Exceptions

        $this->setBaseDir($baseDir);
        $this->setBaseUrl($baseUrl);
        $this->setEmail($email); // email address in case of errors
        $this->setRequiredLogin($requiredLogin);
        $this->setAllConditions($allConditions);
        $this->setBlockRandom($blockRandom);
        $this->setPageOrder($pageOrder);
        
        $this->setOutputFile($this->getBaseDir().'/'.self::OUTPUT_FILENAME); // location of the output file
        $this->setPassFile($this->getBaseDir().'/'.self::PASSWORD_FILENAME); // location of the password file
        $this->setBlockRandomFile($this->getBaseDir().'/'.self::BLOCKRAND_FILENAME); // location of the block randomization file
        
        $this->init();
    }

    /* GETTERS */
    public function getBaseDir() {
        return $this->baseDir;
    }

    public function getBaseUrl() {
        return $this->baseUrl;
    }

    public function getEmail() {
        return $this->email;
    }

    public function getRequiredLogin() {
        return $this->requiredLogin;
    }

    public function getAllConditions() {
        return $this->allConditions;
    }

    public function getBlockRandom() {
        return $this->blockRandom;
    }
    
    public function getPageOrder() {
        return $this->pageOrder;
    }

    public function getOutputFile() {
        return $this->outputFile;
    }

    public function getPassFile() {
        return $this->passFile;
    }

    public function getBlockRandomFile() {
        return $this->blockRandomFile;
    }

    public function getAllVariables() {
        return $this->allVariables;
    }

    public function getSonaID() {
        return $this->sonaID;
    }

    public function getSID() {
        return $this->sID;
    }
    
    public function getOtherID() {
        return $this->otherID;
    }

    public function getCondition() {
        return $this->condition;
    }

    public function getLastPage() {
        return $this->lastPage;
    }
    
    public function getCurrentPage() {
        return $this->currentPage;
    }

    public function getExtras() {
        return $this->extras;
    }

    // Returns extras in format [key_1:value_1,key_2:value_2,...,key_n:value_n]
    public function getExtrasString() {
        $return = '';
        $extras = $this->getExtras();
        if ($extras) {
            $return = '[';
            foreach ($extras as $key => $value) {
                $return .= str_replace(array('|',':',','), '-', trim($key)).':'.str_replace(array('|',':',','), '-', trim($value)).',';
            }
            $return = substr($return, 0, -1).']';
        }
        return $return;
    }
    
    // Parses a string in format [key_1:value_1,key_2:value_2,...,key_n:value_n] and returns array
    public function parseExtrasString($string) {
        $extras = array();
        $string = substr($string, 1, -1);
        $extrasPieces = explode(',', $string);
        foreach ($extrasPieces as $chunk) {
            $chunkPiece = explode(':', $chunk);
            $extras[$chunkPiece[0]] = $chunkPiece[1];
        }
        return $extras;
    }

    /* SETTERS */
    public function setBaseDir($input) {
        $this->baseDir = $input;
        return $this; // allow method chaining
    }

    public function setBaseUrl($input) {
        $this->baseUrl = $input;
        return $this; // allow method chaining
    }

    public function setEmail($input) {
        $this->email = $input;
        return $this; // allow method chaining
    }

    public function setRequiredLogin($input) {
        if ($input == 'sona' || $input == 'password' || $input == 'none') {
            $this->requiredLogin = $input;
        } else {
            $this->requiredLogin = 'none';
        }
        return $this; // allow method chaining
    }

    public function setAllConditions($input) {
        if (is_array($input)) {
            $this->allConditions = $input;
        } elseif (is_string($input) || is_int($input)) {
            $this->allConditions = array(''.$input);
        }
        return $this; // allow method chaining
    }

    public function setBlockRandom($input) {
        if ($input === true || $input === 'true') {
            $this->blockRandom = true;
        } else {
            $this->blockRandom = false;
        }
        return $this; // allow method chaining
    }
    
    public function setPageOrder($input) {
        if (is_array($input)) {
            $this->pageOrder = $input;
        }
        
        return $this; // allow method chaining
    }

    public function setOutputFile($input) {
        $this->outputFile = $input;
        return $this; // allow method chaining
    }

    public function setPassFile($input) {
        $this->passFile = $input;
        return $this; // allow method chaining
    }

    public function setBlockRandomFile($input) {
        $this->blockRandomFile = $input;
        return $this; // allow method chaining
    }

    public function setAllVariables($input) {
        $this->allVariables = $input;
        return $this; // allow method chaining
    }

    public function setSonaID($input) {
        $this->sonaID = $input;
        return $this; // allow method chaining
    }

    public function setSID($input) {
        $this->sID = $input;
        return $this; // allow method chaining 
    }
    
    public function setOtherID($input) {
        $this->otherID = $input;
        return $this; // allow method chaining 
    }

    public function setCondition($input) {
        $this->condition = $input;
        return $this; // allow method chaining
    }

    public function setLastPage($input) {
        $this->lastPage = $input;
        return $this; // allow method chaining
    }
    
    public function setCurrentPage($input) {
        $this->currentPage = $input;
        return $this; // allow method chaining
    }

    public function setExtras($input) {
        $this->extras = $input;
        return $this; // allow method chaining
    }

    /**
     * Function to handle exceptions gracefully
     * @param array $exception The exception that has been thrown
     */
    public static function exceptionHandler($exception) {
        print '<!DOCTYPE html>
<html>
    <head>
        <title>Error</title>
    </head>
    <body>
        <strong>Error:</strong> '.$exception->getMessage().'
    </body>
</html>';
    }
    
    /**
     * Sets participant-specific variables, checks login, and then displays the correct page
     */
    protected function init() {
        // get or choose condition
        if (!empty($_POST[self::FORM_PREFIX.'cond'])) {
            $this->setCondition(explode('|', $_POST[self::FORM_PREFIX.'cond']));
        } else {
            $this->chooseCondition();
        }
        
        // set page information
        $prevPage = (!empty($_POST[self::FORM_PREFIX.'thisPage'])) ? $_POST[self::FORM_PREFIX.'thisPage'] : null;
        $this->setLastPage($prevPage);
        $this->setCurrentPage($this->nextPage($prevPage));
        
        // check for consent
        // CONSTRAINT: consent page must have a field named 'consent' or 'vv0001' with values 'Accept' and 'Decline'
        if (substr($this->getLastPage(), 0, 7) == 'consent') {
            $consent = (isset($_POST['consent'])) ? $_POST['consent'] : $_POST['vv0001'];
            if (!empty($consent)) {
                if ($consent == 'Decline') {
                    $this->setCurrentPage('declined.html');
                } elseif($consent != 'Accept') {
                    throw new Exception('You must indicate whether you accept or decline to participate in this study. Please press the "Back" button and respond to this question.');
                }
            } else {
                throw new Exception('You must indicate whether you accept or decline to participate in this study. Please press the "Back" button and respond to this question.');
            }
        }

        $this->checkLogin($this->getRequiredLogin());

        // set extra variables passed through $_GET
        $extras = array();
        if (!empty($_POST[self::FORM_PREFIX.'extras'])) {
            $extras = $this->parseExtrasString($_POST[self::FORM_PREFIX.'extras']);
        }
        if (count($_GET) > 0) {
            foreach ($_GET as $key => $value) {
                if (!isset($extras[$key])) {
                    $extras[$key] = $value;
                }
            }
        }
        if (!empty($extras)) {
            $this->setExtras($extras);
        }

        $this->setAllVariables($this->findAllVariables());

        $this->saveData($_POST);
        $this->chmodFiles();
        
        $this->showFile($this->getCurrentPage());
    }

    /**
     * Chooses one level of each independent variable to specify one unique condition for a participant
     */
    protected function chooseCondition() {
        $conditions = $this->getAllConditions();
        
        if (count($conditions) > 0) {
            $blockRandom = $this->getBlockRandom();

            $file = $this->getBlockRandomFile();
            $flag = 'a'; // open for writing

            $condKeys = array();
            $this->allConditionKeys($conditions, $condKeys); // get all possible combinations
            $condKeysTemp = $condKeys;

            $condCount = count($condKeys);

            if (($blockRandom == true) && file_exists($file)) {
                // if a condition has already been used, unset it from the list
                $lines = file($file, FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES);
                foreach ($lines as $line) {
                    if (($key = array_search($line, $condKeys)) !== false) {
                        unset($condKeys[$key]);
                    }
                }
                $condKeys = array_values($condKeys); // reset array keys
                
                $condCount = count($condKeys);
                if ($condCount == 0) {
                    $condKeys = $condKeysTemp; // reset block randomization
                    $condCount = count($condKeys);
                    $flag = 'w'; // open for writing, truncate file to zero
                }
            }

            // select condition
            $rand = mt_rand(0, ($condCount-1));
            $selectedKey = $condKeys[$rand];

            if ($blockRandom == true) {
                $handle = fopen($file, $flag);
                fwrite($handle, $selectedKey."\n");
                fclose($handle);
            }

            $finalCondition = array();
            $selectLength = strlen($selectedKey);
            for ($i = 0; $i < $selectLength; $i++) {
                $finalCondition[] = $conditions[$i][(substr($selectedKey, $i, 1))];
            }

            $this->setCondition($finalCondition);
        } else {
            $this->setCondition(array());
        }
    }
    
    /**
     * When given a page name, will find the next page that exists in the correct order
     * @param string $prevPage The name of the page for which to find the next page (optional, defaults to current page)
     * @return string Returns the name of the next page
     */
    protected function nextPage($prevPage = null) {
        if ($prevPage === null) {
            $prevPage = $this->getLastPage();
        }
    
        // try up to 30 times to find a page to display
        for ($counter = 0; $counter < 30; $counter++) {
            $pageOrder = $this->getPageOrder();

            // if user has specified page order
            if ($pageOrder !== null) {
                if ($prevPage === null) {
                    $check = $pageOrder[0];
                } else {
                    $prevPageDetails = $this->parseFilename($prevPage);
                    $key = array_search($prevPageDetails['name'], $pageOrder);
                    if ($key !== false) {
                        $check = $pageOrder[($key+1)];
                    } else {
                        $check = '';
                    }
                }

            // else, use default page order
            } else {
                if ($prevPage == null) {
                    $check = 'index';
                } elseif (substr($prevPage, 0, 5) == 'index') {
                    $check = 'consent';
                } elseif (substr($prevPage, 0, 7) == 'consent') {
                    $check = 'page1';
                } elseif(substr($prevPage, 0, 4) == 'page') {
                    $check = 'page'.(substr($prevPage, 4) + 1);
                }
            }
            
            // check for condition-specific pages
            $condCombinations = $this->condComb($this->getCondition());
            $condsCount = count($condCombinations);
            for ($i = 0; $i < $condsCount; $i++) {
                $condCheck = $check.'_'.$condCombinations[$i].'.html';
                if (file_exists($this->getBaseDir().'/'.$condCheck)) {
                    return $condCheck;
                }
            }
            
            // otherwise, check for regular page
            if (file_exists($this->getBaseDir().'/'.$check.'.html')) {
                return $check.'.html';
            } else {
                if ($pageOrder === null) {
                    $prevPage = $check.'.html'; // if page cannot be found, reset and check for the next one in sequence
                } else {
                    throw new Exception('Page could not be found. Try refreshing the page, and if this error appears again, please contact '.$this->getEmail().' (code 2)');
                }
            }
        }
        
        // if no page found
        if (file_exists($this->getBaseDir().'/thanks.html')) { // one more page to check
            return 'thanks.html';
        } else {
            throw new Exception('Page could not be found. Try refreshing the page, and if this error appears again, please contact '.$this->getEmail().' (code 2)');
        }
    }

    /**
     * If login is required (whether through Sona or a password), checks to make sure participant has logged in
     * @param string $requiredLogin The type of login that is required by the survey
     * @return boolean Returns true if login was successful
     */
    protected function checkLogin($requiredLogin) {
        // set Sona ID
        if ($requiredLogin == 'sona') {
            if (!empty($_GET['sonaid']) && $_GET['sonaid'] != '%SURVEY_CODE%') {
                $this->setSonaID(preg_replace('/[^a-zA-Z0-9]/', '-', $_GET['sonaid'])); // replace non-alphanumeric characters with '-'
                unset($_GET['sonaid']);
                return true;
            } elseif (!empty($_POST[self::FORM_PREFIX.'sonaID']) && $_POST[self::FORM_PREFIX.'sonaID'] != '---SONAID---') {
                $this->setSonaID($_POST[self::FORM_PREFIX.'sonaID']);
                return true;
            } else {
                throw new Exception('Please login directly from the SONA web site. If you continue to have problems accessing the study, please contact '.$this->getEmail().' (code 3)');
            }

        // set sID
        } elseif ($requiredLogin == 'password') {
            if(substr($this->getCurrentPage(), 0, 5) != 'index'
                && substr($this->getCurrentPage(), 0, 7) != 'consent'
                && substr($this->getCurrentPage(), 0, 8) != 'declined'
            ) {
                if (!empty($_POST[self::FORM_PREFIX.'sID'])) {
                    $this->setSID($_POST[self::FORM_PREFIX.'sID']);
                    return true;
                } else {
                    if (!file_exists($this->getPassFile())) {
                        throw new Exception('Cannot access the username and password information. If you continue to have problems accessing the study, please contact '.$this->getEmail().' (code 4)');
                    }
                    
                    // CONSTRAINT: consent page must have fields named 'sID' and 'password'
                    if (empty($_POST['sID']) || empty($_POST['password'])) {
                        throw new Exception('Please login with the username and password provided to you. If you continue to have problems accessing the study, please contact '.$this->getEmail().' (code 3)');
                    }

                    $idInfo = $this->checkUserAndPass($_POST['sID'], $_POST['password']);
                    if ($idInfo == null) {
                        throw new Exception('Please login with the username and password provided to you. If you are having problems accessing the study, please contact '.$this->getEmail().' (code 3)');
                    } else {
                        $this->setSID($idInfo['sID']);
                        $this->setSonaID($idInfo['mtID']);
                        unset($_POST['sID']);
                        unset($_POST['password']);
                        return true;
                    }
                }
            }

        // for when no participant identifier is used
        } else {
            if (!empty($_POST[self::FORM_PREFIX.'sID'])) {
                $this->setSID($_POST[self::FORM_PREFIX.'sID']);
                return true;
            } else {
                $otherID = date('Ymd-His').'_'.mt_rand(10000, 99999);
                $this->setSID($otherID);
            }
            return true;
        }
    }

    /**
     * Checks the given username and password with existing usernames and passwords in passfile.txt
     * @param string $user The username to be checked
     * @param string $pass The password to be checked
     * @return array Returns array with username and Mass Testing ID (if present in file)
     */
    protected function checkUserAndPass($user, $pass) {
        $return = null;

        $idList = file($this->getPassFile(), FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES);
        $match = false;

        foreach ($idList as $line) {
            //$pieces = explode("\s", $line);
            $pieces = preg_split("/[\s]+/", $line);
            if ($user == $pieces[0] && $pass == $pieces[1]) {
                $match = true;
                $return['sID'] = $pieces[0];
                $return['mtID'] = (isset($pieces[2])) ? $pieces[2] : null;
                break;
            }
        }
        return $return;
    }

    /**
     * Gets all variables from output.txt, if stored, or else searches all HTML files for variables and stores results in output.txt
     * @return array Returns an array of all variable names, separated by page name
     */
    protected function findAllVariables() {
        $outputFile = $this->getOutputFile();
        if (file_exists($outputFile)) {
            $handle = fopen($outputFile, 'r'); // open for reading only, with pointer at beginning of file
            $firstLine = rtrim(fgets($handle)); // get just first line of file
            fclose($handle);

            $variables = json_decode($firstLine, true); // returns null if not a valid JSON string
        } else {
            $variables = null;
        }

        // if variables cannot be found (no output file or incorrect JSON format), go through all HTML files and find variables
        if ($variables == null) {
            $variables = $this->findVariablesinHTML();
            $save = json_encode($variables)."\n";

            $handle = fopen($outputFile, 'c'); // open for writing only, with pointer at beginning of file; create file if it does not exist
            if (fwrite($handle, $save) === false) {
                throw new Exception('There was an issue with saving your data. If this error continues to appear, please contact '.$this->getEmail().' (code 5)');
            }
            fclose($handle);
        }
        return $variables;
    }

    /**
     * Searches all HTML files and gets variable names of form elements
     * @return array Returns an array of all variable names, separated by page name
     */
    protected function findVariablesinHTML() {
        $baseDir = $this->getBaseDir();

        //$pageFiles = array();
        //$pageNames = array();
        $variables = array();

        if ($dh = opendir($baseDir)) {
            while (($file = readdir($dh)) !== false) {
                if ($file !== '.' && $file !== '..' && !is_dir($file)
                    && (substr($file, -5) == '.html' || substr($file, -4) == '.htm')
                ) {
                    $contents = file_get_contents($baseDir.'/'.$file);

                    // find all HTML form elements and collect names
                    if (preg_match_all('/<(?:input|select|textarea)[^>]*name=(?:"([^"]+)"|\'([^\']+)\')[^>]*>/', $contents, $matches)) {
                        $pageDetails = $this->parseFilename($file);
                        $arrayKey = $pageDetails['fullName'];
                        //$pageFiles[] = $arrayKey;
                        //$pageNames[] = $pageDetails['name'];
                        $variables[$arrayKey] = array();
                        $matchesCount = count($matches[0]);
                        for ($i = 0; $i < $matchesCount; $i++) {
                            $variables[$arrayKey][$i] = (!empty($matches[1][$i])) ? $matches[1][$i] : $matches[2][$i]; // get variable in either single or double quotes

                            // just a couple variables that we don't want to keep
                            if ($pageDetails['name'] == 'consent'
                                && ($variables[$arrayKey][$i] == 'sID' || $variables[$arrayKey][$i] == 'password')
                            ) {
                                unset($variables[$arrayKey][$i]);
                            }
                        }
                        $variables[$arrayKey] = array_values(array_unique($variables[$arrayKey])); // remove duplicates (like from sets of radio buttons)
                    }
                }
            }

            /*$pageNameDuplicates = uniqueValues($pageNames); // get pages with condition-specific information
            foreach ($pageNameDuplicates as $file => $keys) {
                $variablesTemp = array();
                foreach ($keys as $key) {
                    $fileName = $pageFiles[$key];
                    $variablesTemp[$fileName] = $variables[$fileName];
                }
                $variablesIntersect = call_user_func_array('array_intersect', $variablesTemp); // get variables common to all conditions
                foreach ($variablesTemp as $fileName => $vars) {
                    $variableUnique = array_diff($variablesTemp[$fileName], $variablesIntersect); // get variables unique to a condition
                    // remove common variables from each condition
                    if (count($variableUnique) > 0) {
                        $variables[$fileName] = $variableUnique;
                    } else {
                        unset($variables[$fileName]);
                    }
                }
                $variables[$file] = $variablesIntersect;
            }
            ksort($variables);*/

            return $variables;

        } else {
            throw new Exception('Could not open directory to read files. If this error continues to appear, please contact '.$this->getEmail().' (code 5)');
        }
    }
    
    /**
     * Saves participant-submitted data in the output file
     * @param array $postData The array of submitted form values
     */
    protected function saveData($postData) {
        $outputVars = array();
        $countVars = 0;
        $output = '';

        // set some initial variables (passed through hidden input fields)
        $initialVars = array(
            array($this->getSID(), self::FORM_PREFIX.'sID'),
            array($this->getSonaID(), self::FORM_PREFIX.'sonaID'),
            array($this->getLastPage(), self::FORM_PREFIX.'thisPage'),
            array(implode('*', $this->getCondition()), self::FORM_PREFIX.'cond'),
            array($this->getExtrasString(), self::FORM_PREFIX.'extras')
        );
        foreach ($initialVars as $key => $variable) {
            if ($variable[0]) {
                $outputVars[$key] = $variable[0];
            } else {
                $outputVars[$key] = (!empty($postData[($variable[1])])) ? $postData[($variable[1])] : '';
            }
            unset($postData[($variable[1])]); // unset so it doesn't get used later
        }
        if ($outputVars[0] == '') {
            $outputVars[0] = $this->getOtherID();
        }
        
        // get variables that are accounted for
        $allVariables = $this->getAllVariables(); // gets list of all variables in all files
        $pageDetails = $this->parseFilename($outputVars[2]);
        $page = $pageDetails['fullName'];
        if (isset($allVariables[$page]) && count($allVariables[$page]) > 0) {
            foreach ($allVariables[$page] as $var) {
                if (isset($postData[$var])) {
                    $output .= '|'.str_replace(array("\r\n", "\n", "\r"), '<br>', trim($postData[$var])); // convert newlines to <br>
                    unset($postData[$var]);
                } else {
                    $output .= '|'; // leave blank space for empty variables
                }
            }
            $countVars = count($allVariables[$page]);
        }

        // if there are extra variables not accounted for, insert them at the end
        if (count($postData) > 0) {
            foreach ($postData as $name => $var) {
                // just a couple variables that we don't want to keep
                if (!($pageDetails['name'] == 'consent' && ($name == 'sID' || $name == 'password'))) {
                    $output .= '|'.str_replace(array("\r\n", "\n", "\r"), '<br>', trim($var)); // convert newlines to <br>
                    $countVars++;
                }
            }
        }
    
        // if there are any variables to store, set up the output to be stored
        if ($countVars > 0) {
            $outputVars[6] = date('Y-m-d|H:i:s'); // sets two variables!
            $outputVars[7] = $countVars;
            $outputVars[8] = 'qqstart';

            $output = implode('|', $outputVars).$output."\n"; // final string

            $handle = fopen($this->getOutputFile(), 'a'); // open for writing only, with pointer at end of file; create file if it does not exist
            if (fwrite($handle, $output) === false) {
                throw new Exception('There was an issue with saving your data. If this error continues to appear, please contact '.$this->getEmail().' (code 1)');
            }
            fclose($handle);
        }
    }
    
    /**
     * Ensures that all survey files are protected from external parties
     */
    protected function chmodFiles() {
        $baseDir = $this->getBaseDir();
        if ($dh = opendir($baseDir)) {
            while (($file = readdir($dh)) !== false) {
                if ($file !== '.' && $file !== '..' && !is_dir($file)
                    && (substr($file, -4) == '.txt' || substr($file, -5) == '.html' || substr($file, -4) == '.htm')
                ) {
                    $currentFile = $baseDir.'/'.$file;
                    chmod($currentFile, 0600); // inaccessible for everyone except Owner
                }
            }
        }
    }
    
    /**
     * Reads contents of a page, replaces inserted elements, and then displays the page
     * @param string $page The filename of the page to be displayed (without filepath)
     */
    protected function showFile($page) {
        $contents = file_get_contents($this->getBaseDir().'/'.$page);
        $contents = $this->insertHidden($contents);
        $contents = $this->replaceInserts($contents);
        print $contents;
    }

    /**
     * Inserts hidden form input elements in order to carry important data from one page to another
     * @param string $file The text contents of a page on which to insert the hidden elements
     * @return string Returns the modified text contents
     */
    protected function insertHidden($file) {
        $inputs = '<input type="hidden" name="'.self::FORM_PREFIX.'sonaID" value="'.$this->getSonaID().'" />
<input type="hidden" name="'.self::FORM_PREFIX.'sID" value="'.$this->getSID().'" />
<input type="hidden" name="'.self::FORM_PREFIX.'cond" value="'.implode('|', $this->getCondition()).'" />
<input type="hidden" name="'.self::FORM_PREFIX.'thisPage" value="'.$this->getCurrentPage().'" />
<input type="hidden" name="'.self::FORM_PREFIX.'extras" value="'.$this->getExtrasString().'" />'."\n";
        
        $file = preg_replace('/(<form[^>]*>)/', '$1'.$inputs, $file); // inserts hidden inputs after every form tag
        
        /*$formStart = strpos($file, '<form'); // finds beginning of opening form tag
        if ($formStart !== false) {
            $formEnd = strpos($file, '>', $formStart); // finds end of opening form tag
            $file = substr_replace($file, $inputs, $formEnd+1, 0);
        }*/
        return $file;
    }
    
    /**
     * Checks for inserts, like condition-specific text and responses from previous pages, and replaces them with the proper details
     * @param string $file The text contents of a page for which to replace inserts
     * @return string Returns the modified text contents
     */
    protected function replaceInserts($file) {
        $fileLen = strlen($file);
        $pos = $fileLen-1;
        while ($pos > 0) { // start from end and work backwards, to allow for nested inserts
            $insertBeg = strrpos(substr($file, 0, $pos), '[%'); // finds beginning of an insert
            if ($insertBeg !== false) {
                $insertEnd = strpos($file, '%]', $insertBeg);
                $newContent = $this->parseInsert(substr($file, $insertBeg, ($insertEnd-$insertBeg+2)));
                $file = substr_replace($file, $newContent, $insertBeg, ($insertEnd-$insertBeg+2));
                
                $pos = $insertBeg-1;
            } else {
                $pos = -1;
            }
        }
        return $file;
    }

    /**
     * Parses the content of one insert tag, and returns the new content to replace it
     * @param string $insert The text of the insert tag
     * @return string Returns the new content to replace the insert tag
     */
    protected function parseInsert($insert) {
        $output = '';
        $insertNameEnd = strpos($insert, ':', 2);
        $insertName = substr($insert, 2, $insertNameEnd-2);
        
        if ($insertName == 'include') {
            $output = $this->parseIncludeInsert(substr($insert, $insertNameEnd+1, -2));
        } elseif($insertName == 'condition') {
            $output = $this->parseConditionInsert(substr($insert, $insertNameEnd+1, -2));
        }

        return $output;
    }

    /**
     * Replaces an 'include' insert with the participant's response from a previous page
     * @param string $insertContent The text of the insert tag, after the name of the tag
     * @return string Returns the modified text contents
     */
    protected function parseIncludeInsert($insertContent) {
        $output = '';
        if (preg_match('/(?:\'([^\']+)\'|"([^"]+)"),(?:\'([^\']+)\'|"([^"]+)")/', $insertContent, $matches)) {
            $page = (!empty($matches[1])) ? $matches[1] : $matches[2]; // get page in either single or double quotes
            $fullPage = $page;
            $varID = (!empty($matches[3])) ? $matches[3] : $matches[4];
            
            // pulling a variable from $_GET rather than a specific page
            if ($page == 'url_variable') {
                if ($varID == 'sonaid') {
                    $output = $this->getSonaID();
                } else {
                    $extras = $this->getExtras();
                    if (isset($extras[$varID])) {
                        $output = $extras[$varID];
                    }
                }
            
            // normal include from a page variable
            } else {

                // look for condition-specific page variables first
                $allVariables = $this->getAllVariables();
                $condCombinations = $this->condComb($this->getCondition());
                $condsCount = count($condCombinations);
                for ($i = 0; $i < $condsCount; $i++) {
                    if (array_key_exists($page.'_'.$condCombinations[$i], $allVariables)) {
                        $fullPage = $page.'_'.$condCombinations[$i];
                        break;
                    }
                }

                // convert variable name into variable number if necessary
                if (is_numeric($varID)) {
                    $varNum = $varID;
                } else {
                    $varNum = (array_search($varID, $allVariables[$fullPage]) !== false)
                        ? array_search($varID, $allVariables[$fullPage]) : false;
                }

                // figure out which ID to check for
                if ($this->getSonaID()) {
                    $id = $this->getSonaID();
                    $pieceCheck = 1;
                } elseif($this->getSID()) {
                    $id = $this->getSID();
                    $pieceCheck = 0;
                }

                // open output.txt, find sonaID or sID, then find page, then find variable
                if (!empty($page) && $varNum !== false && !empty($id) && file_exists($this->getOutputFile())) {
                    $lines = file($this->getOutputFile(), FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES);
                    for ($i = (count($lines)-1); $i >= 0; $i--) {
                        $pieces = explode('|', $lines[$i]);

                        $outputPage = $this->parseFilename($pieces[2]);
                        if ($pieces[$pieceCheck] == $id && $outputPage['name'] == $page) {
                            $output = $pieces[$varNum+9]; // add 9 to get to the start of actual submitted answers
                            break;
                        }
                    }
                }
            
            }
        }

        return $output;
    }
    
    /**
     * Replaces a 'condition' insert with condition-specific content
     * @param string $insertContent The text of the insert tag, after the name of the tag
     * @return string Returns the modified text contents
     */
    protected function parseConditionInsert($insertContent) {
        $output = '';
        if (preg_match_all('/(?:\'([^\']+)\'|"([^"]+)"):{([^}]*)}/', $insertContent, $matches)) {
            $conditions = array();
            $text = array();
            $matchesCount = count($matches[0]);
            for ($i = 0; $i < $matchesCount; $i++) {
                $conditions[$i] = (!empty($matches[1][$i])) ? $matches[1][$i] : $matches[2][$i]; // get condition in either single or double quotes
                $text[$i] = $matches[3][$i];
            }

            $condCombinations = $this->condComb($this->getCondition());
            $condsCount = count($condCombinations);
            for ($i = 0; $i < $condsCount; $i++) {
                $key = array_search($condCombinations[$i], $conditions);
                if ($key !== false) {
                    $output = $text[$key];
                    break;
                }
            }
        }

        return $output;
    }
    
    /**
     * Splits up a filename into its individual elements
     * @param string $file The filename to be parsed
     * @return array Returns an array with the individual elements of the filename: name, condition (if applicable), full name (name and condition), and extension
     */
    protected function parseFilename($file) {
        $dotPieces = explode('.', $file);
        $extension = array_pop($dotPieces); // stuff after last dot is file extension
        $filename = implode('.', $dotPieces);
        
        $underPieces = explode('_', $filename);
        $underPiecesCount = count($underPieces);
        if ($underPiecesCount > 1) {
            $condition = array_pop($underPieces);
        } else {
            $condition = '';
        }
        $name = implode('_', $underPieces);
        
        return array('name' => $name, 'condition' => $condition, 'fullName' => $filename, 'extension' => $extension);
    }
    
    /**
     * Creates an array with all combinations of condition names, from most specific (i.e. 'abc') to least specific (i.e. 'a')
     * @param array $conds An array specifying the conditions in which the participant is
     * @return array Returns an array with all combinations of condition names
     */
    protected function condComb($conds) {
        $length = count($conds);
        if ($length > 0) {
            $combocount = pow(2, $length);
            for ($i = 1; $i < $combocount; $i++) {
                $binary[($i-1)] = str_pad(decbin($i), $length, "0", STR_PAD_LEFT); // create all combinations of binary choices (1 is present, 0 is absent)
                $combination = '';
                for ($j = 0; $j < $length; $j++) {
                    if ($binary[($i-1)][$j] == "1") {
                        $combination .= $conds[$j]; // string together all 'present' array elements
                    }
                }
                $combsArray[] = $combination;
            }

            uasort($binary, array(get_class(), 'sortBinary')); // custom function to sort binary choices how I want them
            $keys = array_keys($binary); // get order from sort
            foreach ($keys as $key) { // sort combinations
                $conditionsArray[] = $combsArray[$key];
            }
            return $conditionsArray;
        } else {
            return null;
        }
    }

    /**
     * Custom sort function; sorts by number of 1s, then by magnitude; works in conjunction with condComb()
     * @param string $a First item to be compared
     * @param string $b Second item to be compared
     * @return int Returns an integer indicating sort order; if $a > $b, returns negative; if $a < $b, returns positive; if $a == $b, returns 0
     */
    protected static function sortBinary($a, $b) {
        // sort by number of binary 1s first (length)
        if ((substr_count($b, '1')-substr_count($a, '1')) == 0) {
            // sort by magnitude after that
            return bindec($b)-bindec($a);
        } else {
            return substr_count($b, '1')-substr_count($a, '1');
        }
    }

    /**
     * Creates a list of all possible combinations of the levels of the independent variables (by array key)
     * @param array $conds A multi-dimensional array of all the conditions
     * @param array &$output A new array into which to put the combinations
     */
    protected function allConditionKeys($conds, &$output, $string = '', $i = 0) {
        // check if array is multi-dimensional
        if (count($conds) == count($conds, COUNT_RECURSIVE)) {
            $output = array_keys($conds);
        } else {
            if ($i >= count($conds)) {
                $output[] = $string;
            } else {
                if (is_array($conds[$i])) {
                    foreach ($conds[$i] as $key => $trait) {
                        $this->allConditionKeys($conds, $output, $string.$key, $i+1);
                    }
                } else {
                    $this->allConditionKeys($conds, $output, $string.'0', $i+1);
                }
            }
        }
    }
}
?>